import vcf
import pyranges
from .json_generator import _get_fhir_json
import logging
import sys
general_logger = logging.getLogger('vcf2fhir.general')

class Converter(object):
    """Converter for a VCF v >4.1 file."""
    def __init__(self, vcf_filename=None, ref_build=None, patient_id = None, has_tabix=False, conv_region_filename=None, conv_region_dict = None, region_studied_filename= None, nocall_filename = None):
        """ 
        Create a new Converter Object to convert a VCF file.

        Parameters
        ----------
        vcf_filename : str (Required)
            Path to text-based or bgzipped VCF file containing variants to be converted into FHIR format.
            Valid path and filename without whitespace. VCF file must conform to VCF Version 4.1 or later. 
            FORMAT.GT must be present. Multi-sample VCFs are allowed, but only the first sample will be converted.
        ref_build : str (Required)
            Genome Reference Consortium genome assembly to which variants in the VCF were called. Must be one of 'GRCh37' or 'GRCh38'.
        patient_id : str (Optional)
            Patient who's VCF file is being processed. Alphanumeric string without whitespace. Default value is first sample name.
        has_tabix : bool (Optional)
            If tabix file exist for the vcf than set it to True. Tabix file should have the same name as vcf file, with a '.tbi' extension,
            and must be in the same folder. Default value is False.
        conv_region_filename : str (Optional)
            Path to conversion region bed file. Subset of the VCF file to be converted into FHIR. If absent, the entire VCF file is converted. Must be a valid BED file
        conv_region_dict: dict (Optional)
            Conversion region can also be provided using dict. If 'conv_region_filename' is provided
            it will be ignored. 
            Format:  {"Chromosome": ["chr1", "chr2"], "Start": [100, 200], "End": [150, 201]}
        region_studied_filename : str (Optional)
            Path to region studied bed file. Subset of patient's genome that was studied in the generation of the VCF file. If present, only studied regions are converted. Must be a valid BED file, with first 3 columns: <chr> <start> <stop>.
        nocall_filename : str (Optional)
            Path to no call bed file. Subset of studied region that is deemed noncallable. If present, only studied regions minus noncallable regions are converted. Must be a valid BED file, with first 3 columns: <chr> <start> <stop>.

        Returns
        -------
        Object
        An Instance of Conveter that helps to convert vcf file.

        Examples
        --------

        """
        super(Converter, self).__init__()
        if not (vcf_filename):
            raise Exception('You must provide vcf_filename')
        if not ref_build or ref_build not in ["GRCh37", "GRCh38"]:
            raise Exception('You must provide build number ("GRCh37" or "GRCh38")')
        if nocall_filename and not region_studied_filename:
            raise Exception ("Please also provide region_studied_filename when nocall_filename is provided")
        self.vcf_filename = vcf_filename
        try:
            self._vcf_reader = vcf.Reader(filename=vcf_filename)
        except FileNotFoundError:
            raise
        except :
            self._generate_exception("Please provide valid  'vcf_filename'")
        if not patient_id:
            patient_id = self._vcf_reader.samples[0]
        if nocall_filename:
            try:
                self.nocall_region = pyranges.read_bed(nocall_filename)
            except FileNotFoundError:
                raise
            except:
                self._generate_exception("Please provide valid  'nocall_filename'")
        else:
            self.nocall_region = pyranges.PyRanges()
        if conv_region_filename:
            try:
                self.conversion_region = pyranges.read_bed(conv_region_filename)
            except FileNotFoundError:
                raise
            except:
                self._generate_exception( "Please provide valid 'conv_region_filename'")
        elif conv_region_dict:      
            try:
                self._fix_conv_region_zero_based(conv_region_dict)
                self.conversion_region = pyranges.from_dict(conv_region_dict)
            except FileNotFoundError:
                raise
            except:
                self._generate_exception("Please provide valid 'conv_region_dict'")
        else:
            self.conversion_region = None         
        if region_studied_filename:
            try:
                self.region_studied = pyranges.read_bed(region_studied_filename)
            except FileNotFoundError:
                raise
            except:
                self._generate_exception("Please provide valid 'region_studied_filename'")
        else:
            self.region_studied = None
        self.has_tabix = has_tabix
        self.patient_id = patient_id
        self.ref_build = ref_build
        self.nocall_filename = nocall_filename
        self.conv_region_filename = conv_region_filename
        general_logger.info("Converter class instantiated successfully")
    def convert(self, output_filename='fhir.json'):
        """ Returns True on success

        Parameters
        ----------
        output_filename : str, default fhir.json
            Path to output fhir json.

        Returns
        ----------
        bool True on successful conversion

        """
        try:
            general_logger.info("Starting VCF to FHIR Conversion")
            _get_fhir_json(self._vcf_reader, self.ref_build, self.patient_id, self.has_tabix, self.conversion_region, self.region_studied, self.nocall_region, output_filename)
        except:
            general_logger.error("Error in converting vcf file", exc_info=True)
            return False
        return True

    def _fix_conv_region_zero_based(self, conv_region_dict):
        i = 0
        for start in conv_region_dict["Start"]:
            conv_region_dict["Start"][i] = start - 1
            i += 1
        i = 0
        for end in conv_region_dict["End"]:
            conv_region_dict["End"][i] = end - 1
            i += 1

    def _generate_exception(self, msg):
        general_logger.error(msg, exc_info=True)
        raise Exception (msg, sys.exc_info)




import doctest
import unittest
import os
import vcf2fhir
import json
from os.path import join, dirname
import shutil
import logging
from vcf2fhir.common import _Utilities

suite = doctest.DocTestSuite(vcf2fhir)

def _get_uids_map(fhir_json):
    obv_ids = []
    result_ids = []
    for index, obv in enumerate(fhir_json['contained']):
        id = obv['id']
        obv_ids.append(f'#{id}')
        fhir_json['contained'][index]['id'] = ''

    for index, result in enumerate(fhir_json['result']):
        result_ids.append(result['reference'])
        fhir_json['result'][index]['reference'] = ''
    fhir_json['id'] = ''
    return {'obv_ids' : obv_ids, 'result_ids' : result_ids}

def _validate_phase_rel(fhir_json, map_variant_index):
    for index_seq, list_index_var in map_variant_index.items():
        for index, ref in enumerate(fhir_json['contained'][index_seq]['derivedFrom']):
            variant_id = fhir_json['contained'][list_index_var[index]]['id']
            if not ref['reference'] == f'#{variant_id}':
                return False
            fhir_json['contained'][index_seq]['derivedFrom'][index]['reference'] = ''
    return True

    
class TestVcf2FhirInputs(unittest.TestCase):  
    
    def test_required_vcf_filename(self):
        with self.assertRaises(Exception) as context:
            vcf2fhir.Converter()
        self.assertTrue('You must provide vcf_filename' in str(context.exception))

    def test_invalid_vcf_filename(self):
        self.assertRaises(FileNotFoundError, vcf2fhir.Converter, *['Hello', 'GRCh38'])

    def test_required_ref_build(self):
        with self.assertRaises(Exception) as context:
            vcf2fhir.Converter(os.path.join(os.path.dirname(__file__), 'vcf_example1.vcf'))
        self.assertEqual('You must provide build number ("GRCh37" or "GRCh38")', str(context.exception))

    def test_invalid_ref_build(self):
        with self.assertRaises(Exception) as context:
            vcf2fhir.Converter(os.path.join(os.path.dirname(__file__), 'vcf_example1.vcf'), 'b38')
        self.assertEqual('You must provide build number ("GRCh37" or "GRCh38")', str(context.exception))
    
    def test_valid_ref_build_37(self):
        oVcf2Fhir = vcf2fhir.Converter(os.path.join(os.path.dirname(__file__), 'vcf_example1.vcf'), 'GRCh37')
        self.assertEqual(type(oVcf2Fhir), vcf2fhir.Converter)

    def test_valid_ref_build_38(self):
        oVcf2Fhir = vcf2fhir.Converter(os.path.join(os.path.dirname(__file__), 'vcf_example1.vcf'), 'GRCh38')
        self.assertEqual(type(oVcf2Fhir), vcf2fhir.Converter)

    def test_conv_region_only(self):
        conv_region_filename = os.path.join(os.path.dirname(__file__),'RegionsToConvert_example3.bed')
        oVcf2Fhir = vcf2fhir.Converter(os.path.join(os.path.dirname(__file__), 'vcf_example3.vcf'), 'GRCh37', 'abc', conv_region_filename =conv_region_filename)
        self.assertEqual(type(oVcf2Fhir), vcf2fhir.Converter)

    def test_conv_region_dict(self):
        conv_region_dict =  {
            "Chromosome": ["X", "X", "M"],
            "Start": [50001, 55001, 50001],
            "End": [52001, 60601, 60026]
        }
        oVcf2Fhir = vcf2fhir.Converter(os.path.join(os.path.dirname(__file__), 'vcf_example3.vcf'), 'GRCh37', 'abc', conv_region_dict=conv_region_dict)
        self.assertEqual(type(oVcf2Fhir), vcf2fhir.Converter)

    def test_conv_region_region_studied(self):        
        region_studied_filename = os.path.join(os.path.dirname(__file__),'RegionsStudied_example3.bed')
        conv_region_filename = os.path.join(os.path.dirname(__file__),'RegionsToConvert_example3.bed')
        oVcf2Fhir = vcf2fhir.Converter(os.path.join(os.path.dirname(__file__), 'vcf_example3.vcf'), 'GRCh37', 'abc', conv_region_filename= conv_region_filename, region_studied_filename=region_studied_filename)
        self.assertEqual(type(oVcf2Fhir), vcf2fhir.Converter)

    def test_conv_region_nocall(self):        
        conv_region_filename = os.path.join(os.path.dirname(__file__),'RegionsToConvert_example3.bed')     
        nocall_filename = os.path.join(os.path.dirname(__file__),'NoncallableRegions_example3.bed')    
        with self.assertRaises(Exception) as context:
            vcf2fhir.Converter(os.path.join(os.path.dirname(__file__), 'vcf_example3.vcf'), 'GRCh37', 'abc',  conv_region_filename =conv_region_filename, nocall_filename=nocall_filename)
        self.assertEqual('Please also provide region_studied_filename when nocall_filename is provided',str(context.exception))

    def test_no_conv_region_region_studied(self):        
        region_studied_filename = os.path.join(os.path.dirname(__file__),'RegionsStudied_example3.bed')        
        oVcf2Fhir = vcf2fhir.Converter(os.path.join(os.path.dirname(__file__), 'vcf_example3.vcf'), 'GRCh37', 'abc', region_studied_filename = region_studied_filename)
        self.assertEqual(type(oVcf2Fhir), vcf2fhir.Converter)

    def test_no_conv_region_nocall(self):
        nocall_filename = os.path.join(os.path.dirname(__file__),'NoncallableRegions_example3.bed') 
        with self.assertRaises(Exception) as context:
            vcf2fhir.Converter(os.path.join(os.path.dirname(__file__), 'vcf_example3.vcf'), 'GRCh37', 'abc', nocall_filename = nocall_filename)
        self.assertEqual('Please also provide region_studied_filename when nocall_filename is provided',str(context.exception))
        

class TestTranslation(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.TEST_RESULT_DIR = os.path.join(os.path.dirname(__file__),'output')        
        if os.path.exists(self.TEST_RESULT_DIR):
            shutil.rmtree(self.TEST_RESULT_DIR)
        os.mkdir(self.TEST_RESULT_DIR)

    @classmethod
    def tearDownClass(self):
        shutil.rmtree(self.TEST_RESULT_DIR)

    def test_wo_patient_id(self):
        self.maxDiff = None
        oVcf2Fhir = vcf2fhir.Converter(os.path.join(os.path.dirname(__file__),'vcf_example1.vcf'), 'GRCh37')
        output_filename = os.path.join(os.path.dirname(__file__), self.TEST_RESULT_DIR,'fhir_wo_patient_example1.json')
        expected_outfult_filename = os.path.join(os.path.dirname(__file__),'expected_example1_wo_patient.json')
        bDone = oVcf2Fhir.convert(output_filename)
        # check if translation was completed
        self.assertEqual(bDone, True)
        actual_fhir_json = json.load(open(output_filename))
        # Validate the pased sequence relationship
        self.assertEqual(_validate_phase_rel(actual_fhir_json, {8 : [5, 6]}), True)
        # Validate: list of observation uids and list of result uids same.
        # Also set the uids to '' to avoid guid comparison in next step
        map_ids = _get_uids_map(actual_fhir_json)
        self.assertEqual(map_ids['obv_ids'], map_ids['result_ids'])
        actual_fhir_json['issued'] = ''
        # Finally, check if the acutal json after removing all uids is same as exppected json
        expected_fhir_json = json.load(open(expected_outfult_filename))
        self.assertEqual(actual_fhir_json, expected_fhir_json)

    def test_with_patient_id(self):
        self.maxDiff = None
        oVcf2Fhir = vcf2fhir.Converter(os.path.join(os.path.dirname(__file__), 'vcf_example1.vcf'), 'GRCh37', 'HG00628')        
        output_filename = os.path.join(os.path.dirname(__file__), self.TEST_RESULT_DIR, 'fhir_with_patient_example1.json')
        expected_outfult_filename = os.path.join(os.path.dirname(__file__),'expected_example1_with_patient.json')
        bDone = oVcf2Fhir.convert(output_filename)
        # check if translation was completed
        self.assertEqual(bDone, True)
        actual_fhir_json = json.load(open(output_filename))
        # Validate the pased sequence relationship
        self.assertEqual(_validate_phase_rel(actual_fhir_json, {8 : [5, 6]}), True)
        # Validate: list of observation uids and list of result uids same.
        # Also set the uids to '' to avoid guid comparison in next step
        map_ids = _get_uids_map(actual_fhir_json)
        self.assertEqual(map_ids['obv_ids'], map_ids['result_ids'])
        actual_fhir_json['issued'] = ''
        # Finally, check if the acutal json after removing all uids is same as exppected json        
        expected_fhir_json = json.load(open(expected_outfult_filename))
        self.assertEqual(actual_fhir_json, expected_fhir_json)

    #FIXME: just a temporary test, later change it to a test that test particular variant
    def test_anotherfile(self):
        oVcf2Fhir = vcf2fhir.Converter(os.path.join(os.path.dirname(__file__), 'vcf_example2.vcf'), 'GRCh37', 'HG00628')
        bDone = oVcf2Fhir.convert(output_filename=os.path.join(os.path.dirname(__file__), self.TEST_RESULT_DIR,'fhir2.json'))
        self.assertEqual(bDone, True)

    def test_region_studied(self):
        self.maxDiff = None
        region_studied_filename = os.path.join(os.path.dirname(__file__),'RegionsStudied_example3.bed')
        conv_region_filename = os.path.join(os.path.dirname(__file__),'RegionsToConvert_example3.bed')
        nocall_filename = os.path.join(os.path.dirname(__file__),'NoncallableRegions_example3.bed')               
        output_filename = os.path.join(os.path.dirname(__file__), self.TEST_RESULT_DIR, 'fhir_example3.json')
        expected_outfult_filename = os.path.join(os.path.dirname(__file__),'expected_example3.json')
        oVcf2Fhir = vcf2fhir.Converter(os.path.join(os.path.dirname(__file__), 'vcf_example3.vcf'), 'GRCh38', 'HG00628', conv_region_filename =conv_region_filename, region_studied_filename=region_studied_filename, nocall_filename=nocall_filename)
        bDone = oVcf2Fhir.convert(output_filename)
        # check if translation was completed
        self.assertEqual(bDone, True)
        actual_fhir_json = json.load(open(output_filename))
        # Validate the pased sequence relationship
        self.assertEqual(_validate_phase_rel(actual_fhir_json, {7: [2, 3], 8 : [3, 4]}), True)
        # Validate: list of observation uids and list of result uids same.
        # Also set the uids to '' to avoid guid comparison in next step
        map_ids = _get_uids_map(actual_fhir_json)
        self.assertEqual(map_ids['obv_ids'], map_ids['result_ids'])
        actual_fhir_json['issued'] = ''
        # Finally, check if the acutal json after removing all uids is same as exppected json        
        expected_fhir_json = json.load(open(expected_outfult_filename))
        self.assertEqual(actual_fhir_json, expected_fhir_json)

    def test_region_studied_dict(self):        
        conv_region_dict =  {
            "Chromosome": ["X", "X", "M"],
            "Start": [50001, 55001, 50001],
            "End": [52001, 60601, 60026]
        }
        self.maxDiff = None
        region_studied_filename = os.path.join(os.path.dirname(__file__),'RegionsStudied_example3.bed')
        nocall_filename = os.path.join(os.path.dirname(__file__),'NoncallableRegions_example3.bed')               
        output_filename = os.path.join(os.path.dirname(__file__), self.TEST_RESULT_DIR, 'fhir_example3_dict.json')
        expected_outfult_filename = os.path.join(os.path.dirname(__file__),'expected_example3.json')
        oVcf2Fhir = vcf2fhir.Converter(os.path.join(os.path.dirname(__file__), 'vcf_example3.vcf'), 'GRCh38', 'HG00628', conv_region_dict= conv_region_dict, region_studied_filename=region_studied_filename, nocall_filename=nocall_filename)
        bDone = oVcf2Fhir.convert(output_filename)
        # check if translation was completed
        self.assertEqual(bDone, True)
        actual_fhir_json = json.load(open(output_filename))
        # Validate the pased sequence relationship
        self.assertEqual(_validate_phase_rel(actual_fhir_json, {7: [2, 3], 8 : [3, 4]}), True)
        # Validate: list of observation uids and list of result uids same.
        # Also set the uids to '' to avoid guid comparison in next step
        map_ids = _get_uids_map(actual_fhir_json)
        self.assertEqual(map_ids['obv_ids'], map_ids['result_ids'])
        actual_fhir_json['issued'] = ''
        # Finally, check if the acutal json after removing all uids is same as exppected json        
        expected_fhir_json = json.load(open(expected_outfult_filename))
        self.assertEqual(actual_fhir_json, expected_fhir_json)

    # Check if region studied observation outside the vcf files are also included in fhir report.  
    def test_multiple_region_studied(self):
        self.maxDiff = None
        region_studied_filename = os.path.join(os.path.dirname(__file__),'RegionsStudied_example4.bed')
        conv_region_filename = os.path.join(os.path.dirname(__file__),'RegionsToConvert_example4.bed')
        nocall_filename = os.path.join(os.path.dirname(__file__),'NoncallableRegions_example4.bed')               
        output_filename = os.path.join(os.path.dirname(__file__), self.TEST_RESULT_DIR, 'fhir_example4.json')
        expected_outfult_filename = os.path.join(os.path.dirname(__file__),'expected_example4.json')
        oVcf2Fhir = vcf2fhir.Converter(os.path.join(os.path.dirname(__file__), 'vcf_example4.vcf'), 'GRCh38', 'HG00628', conv_region_filename =conv_region_filename, region_studied_filename=region_studied_filename, nocall_filename=nocall_filename)
        bDone = oVcf2Fhir.convert(output_filename)
        # check if translation was completed
        self.assertEqual(bDone, True)
        actual_fhir_json = json.load(open(output_filename))
        # Validate the pased sequence relationship
        self.assertEqual(_validate_phase_rel(actual_fhir_json, {31: [24, 25], 32 : [25, 26]}), True)
        # Validate: list of observation uids and list of result uids same.
        # Also set the uids to '' to avoid guid comparison in next step
        map_ids = _get_uids_map(actual_fhir_json)
        self.assertEqual(map_ids['obv_ids'], map_ids['result_ids'])
        actual_fhir_json['issued'] = ''
        # Finally, check if the acutal json after removing all uids is same as exppected json        
        expected_fhir_json = json.load(open(expected_outfult_filename))
        self.assertEqual(actual_fhir_json, expected_fhir_json)

    def test_region_studied_only(self):
        region_studied_filename = os.path.join(os.path.dirname(__file__),'RegionsStudied_example4.bed')             
        output_filename = os.path.join(os.path.dirname(__file__), self.TEST_RESULT_DIR, 'fhir_example4_test.json')
        oVcf2Fhir = vcf2fhir.Converter(os.path.join(os.path.dirname(__file__), 'vcf_example4.vcf'), 'GRCh38', 'HG00628', region_studied_filename=region_studied_filename)
        bDone = oVcf2Fhir.convert(output_filename)
        # check if translation was completed
        self.assertEqual(bDone, True)

    def test_empty_fhir_json(self):
        conv_region_filename = os.path.join(os.path.dirname(__file__),'RegionsToConvert_empty_example4.bed')             
        output_filename = os.path.join(os.path.dirname(__file__), self.TEST_RESULT_DIR, 'fhir_example4_test.json')
        oVcf2Fhir = vcf2fhir.Converter(os.path.join(os.path.dirname(__file__), 'vcf_example4.vcf'), 'GRCh38', 'HG00628', conv_region_filename=conv_region_filename)
        bDone = oVcf2Fhir.convert(output_filename)
        # check if translation was completed
        self.assertEqual(bDone, True)

    def test_tabix(self):
        self.maxDiff = None
        region_studied_filename = os.path.join(os.path.dirname(__file__),'RegionsStudied_example4.bed')
        conv_region_filename = os.path.join(os.path.dirname(__file__),'RegionsToConvert_example4.bed')
        nocall_filename = os.path.join(os.path.dirname(__file__),'NoncallableRegions_example4.bed')         
        output_filename = os.path.join(os.path.dirname(__file__), self.TEST_RESULT_DIR, 'fhir_example4_tabix.json')
        expected_outfult_filename = os.path.join(os.path.dirname(__file__),'expected_example4.json')
        oVcf2Fhir = vcf2fhir.Converter(os.path.join(os.path.dirname(__file__), 'vcf_example4.vcf.gz'), 'GRCh38', 'HG00628', has_tabix= True, conv_region_filename =conv_region_filename, region_studied_filename=region_studied_filename, nocall_filename=nocall_filename)
        bDone = oVcf2Fhir.convert(output_filename)
        # check if translation was completed
        self.assertEqual(bDone, True)
        actual_fhir_json = json.load(open(output_filename))
        # Validate the pased sequence relationship
        self.assertEqual(_validate_phase_rel(actual_fhir_json, {31: [24, 25], 32 : [25, 26]}), True)
        # Validate: list of observation uids and list of result uids same.
        # Also set the uids to '' to avoid guid comparison in next step
        map_ids = _get_uids_map(actual_fhir_json)
        self.assertEqual(map_ids['obv_ids'], map_ids['result_ids'])
        actual_fhir_json['issued'] = ''
        # Finally, check if the acutal json after removing all uids is same as exppected json        
        expected_fhir_json = json.load(open(expected_outfult_filename))
        self.assertEqual(actual_fhir_json, expected_fhir_json)

class TestLogger(unittest.TestCase):
    def setUp(self):
        # create file handler and set level to debug   
        self.log_general_filename = os.path.join(os.path.dirname(__file__), self.LOG_DIR, 'general.log')
        self.log_invalid_record_filename = os.path.join(os.path.dirname(__file__), self.LOG_DIR, 'invalid_record.log')
        genearl_fh = logging.FileHandler(self.log_general_filename)
        invalid_record_fh = logging.FileHandler(self.log_invalid_record_filename)
        genearl_fh.setLevel(logging.DEBUG)
        invalid_record_fh.setLevel(logging.DEBUG)
        # create formatter
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        # add formatter to fh
        genearl_fh.setFormatter(formatter)
        invalid_record_fh.setFormatter(formatter)
        # store as a class variable
        self.genearl_fh = genearl_fh
        self.invalid_record_fh = invalid_record_fh

    @classmethod
    def setUpClass(self):
        self.LOG_DIR = os.path.join(os.path.dirname(__file__),'log')        
        if os.path.exists(self.LOG_DIR):
            shutil.rmtree(self.LOG_DIR)
        os.mkdir(self.LOG_DIR)


    # TODO: Delete the log folder after running all the tests, below method throws error becasue log files are still in use.
    # @classmethod
    # def tearDownClass(self):
    #     shutil.rmtree(self.LOG_DIR)

    def testLoggerWorks(self):        
        region_studied_filename = os.path.join(os.path.dirname(__file__),'RegionsStudied_example3.bed')
        conv_region_filename = os.path.join(os.path.dirname(__file__),'RegionsToConvert_example3.bed')
        nocall_filename = os.path.join(os.path.dirname(__file__),'NoncallableRegions_example3.bed')        
        output_filename = os.path.join(os.path.dirname(__file__), self.LOG_DIR, 'logging_fhir.json')
        # create logger               
        general_logger = logging.getLogger('vcf2fhir.general')
        invalid_record_logger = logging.getLogger('vcf2fhir.invalidrecord') 
        general_logger.setLevel(logging.DEBUG)
        invalid_record_logger.setLevel(logging.DEBUG)
        # add ch to logger        
        general_logger.addHandler(self.genearl_fh)
        invalid_record_logger.addHandler(self.invalid_record_fh)
        oVcf2Fhir = vcf2fhir.Converter(os.path.join(os.path.dirname(__file__), 'vcf_example3.vcf'), 'GRCh38', 'HG00628', conv_region_filename=conv_region_filename, region_studied_filename= region_studied_filename, nocall_filename= nocall_filename)
        oVcf2Fhir.convert(output_filename)
        self.assertEqual(os.path.exists(self.log_general_filename), True)
        self.assertEqual(os.path.exists(self.log_invalid_record_filename), True)

class TestChromIdentifier(unittest.TestCase):
    
    def test_chrom_1_22(self):
        actual_chrom = ['chr1', '1', 'CHR1', '22', 'CHR22', 'chr22']
        expected_chrom = ['1', '1', '1', '22', '22', '22']
        i = 0
        for chrom in actual_chrom:
            self.assertEqual(_Utilities.extract_chrom_identifier(chrom), expected_chrom[i])
            i += 1
    
    def test_chrom_X_Y(self):
        actual_chrom = ['chrX', 'X', 'x', 'CHRX', 'chrY', 'Y', 'Y', 'CHRY']
        expected_chrom = ['X', 'X', 'X', 'X', 'Y', 'Y', 'Y', 'Y']
        i = 0
        for chrom in actual_chrom:
            self.assertEqual(_Utilities.extract_chrom_identifier(chrom), expected_chrom[i])
            i += 1

    def test_chrom_M(self):
        actual_chrom = ['MT', 'm', 'M', 'mt', 'chrm', 'chrM', 'chrmt', 'chrMT' ]
        expected_chrom = ['M', 'M', 'M', 'M', 'M', 'M', 'M', 'M']
        i = 0
        for chrom in actual_chrom:
            self.assertEqual(_Utilities.extract_chrom_identifier(chrom), expected_chrom[i])
            i += 1




        

suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestVcf2FhirInputs))
suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestTranslation))
suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestLogger))
suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestChromIdentifier))

if __name__ == '__main__':
    unittest.main()


import logging
from .gene_ref_seq import _get_ref_seq_by_chrom
from .fhir_helper import _Fhir_Helper
from .common import _Utilities

invalid_record_logger = logging.getLogger("vcf2fhir.invalidrecord")
general_logger = logging.getLogger("vcf2fhir.general")


def _valid_record(record):
    if(record.is_sv == True):
        invalid_record_logger.debug("Reason: VCF INFO.SVTYPE is present. (Structural variants are excluded), Record: %s, considered sample: %s", record, record.samples[0].data)
        return False
    if(record.FILTER is not None and record.FILTER != 'PASS'):
        invalid_record_logger.debug("Reason: VCF FILTER does not equal 'PASS' or '.', Record: %s, considered sample: %s", record, record.samples[0].data)
        return False
    if(len(record.samples) == 0 or record.samples[0].gt_type is None):
        invalid_record_logger.debug("Reason: VCF FORMAT.GT is null ('./.', '.|.', '.', etc), Record: %s, considered sample: %s", record, record.samples[0].data)
        return False
    if(len(record.ALT) != 1 or (record.ALT[0].type != 'SNV' and record.ALT[0].type != 'MNV')):
        invalid_record_logger.debug("Reason: ALT is not a simple character string, comma-separated character string or '.' (Rows where ALT is a token (e.g. 'DEL') are excluded), Record: %s, considered sample: %s", record, record.samples[0].data)
        return False
    if record.CHROM == "M" and ((len(record.samples[0].gt_alleles) == 1 and record.samples[0].gt_alleles[0] == "0") or len(record.samples[0].gt_alleles) == 2):
        invalid_record_logger.debug("Reason: Mitochondrial DNA with GT = 0 or its diploid, Record: %s, considered sample: %s", record, record.samples[0].data)
        return False
    return True  

def _get_chrom(chrom_index):
    if chrom_index == 23:
        return "X"
    elif chrom_index == 24:
        return "Y"
    elif chrom_index == 25:
        return "M"
    return str(chrom_index)

def _fix_regions_chrom(region):
    if region:
        region.Chromosome = region.Chromosome.apply(_Utilities.extract_chrom_identifier)


def _add_record_variants(record, ref_seq, patientID, fhir_helper):
    if(_valid_record(record) == True):
        fhir_helper.add_variant_obv(record, ref_seq)
        
def _add_region_studied(region_studied, nocall_region, fhir_helper, chrom, ref_seq, patientID):
    if((region_studied and not region_studied[chrom].empty) or (nocall_region and not nocall_region[chrom].empty)):
        general_logger.info("Adding region Studied observation for %s", chrom)
        general_logger.debug("Region Examined %s", region_studied[chrom])
        general_logger.debug("Region Uncallable %s", nocall_region[chrom])
        fhir_helper.add_regionstudied_obv(ref_seq, region_studied[chrom], nocall_region[chrom])


def _get_fhir_json(vcf_reader, ref_build, patientID, has_tabix, conversion_region, region_studied, nocall_region, output_filename):
    fhir_helper = _Fhir_Helper(patientID)
    fhir_helper.initalizeReport()
    general_logger.debug("Finished Initializing empty report")
    _fix_regions_chrom(conversion_region)
    _fix_regions_chrom(region_studied)
    _fix_regions_chrom(nocall_region)
    if conversion_region:
        if region_studied:
            region_studied = region_studied.intersect(conversion_region)
            general_logger.debug("Final Conmputed Reportable Query Regions: %s", region_studied)
        if nocall_region:
            nocall_region = nocall_region.intersect(conversion_region)
            general_logger.debug("Final Conmputed NoCall Query Regions: %s", nocall_region)
    general_logger.info("Start adding Region studied observation followed by Variant observation corresponding to that region")
    if has_tabix == True:
        for chrom_index in range(1, 26):
            chrom = _get_chrom(chrom_index)
            ref_seq = _get_ref_seq_by_chrom(ref_build , _Utilities.extract_chrom_identifier(chrom))
            _add_region_studied(region_studied, nocall_region, fhir_helper, chrom, ref_seq, patientID)
            if conversion_region and not conversion_region[chrom].empty:
                for _, row in conversion_region[chrom].df.iterrows():
                    vcf_iterator = None
                    try:
                        vcf_iterator = vcf_reader.fetch(chrom, int(row['Start']), int(row['End']))
                    except ValueError:
                        pass
                    if vcf_iterator:
                        for record in vcf_iterator:
                            record.CHROM = _Utilities.extract_chrom_identifier(record.CHROM)
                            _add_record_variants(record, ref_seq, patientID, fhir_helper)
            elif not conversion_region:
                vcf_iterator = None
                try:
                    vcf_iterator = vcf_reader.fetch(chrom)
                except ValueError:
                    pass
                if vcf_iterator:
                    for record in vcf_iterator:
                        record.CHROM = _Utilities.extract_chrom_identifier(record.CHROM)
                        _add_record_variants(record, ref_seq, patientID, fhir_helper)
    else:
        chrom_index = 1
        prev_add_chrom = ""    
        for record in vcf_reader:
            record.CHROM = _Utilities.extract_chrom_identifier(record.CHROM)
            if not (conversion_region and conversion_region[record.CHROM].empty):
                if prev_add_chrom != record.CHROM and (region_studied or nocall_region):
                    chrom = _get_chrom(chrom_index)
                    while prev_add_chrom != record.CHROM:
                        current_ref_seq = _get_ref_seq_by_chrom(ref_build, chrom)
                        _add_region_studied(region_studied, nocall_region, fhir_helper, chrom, current_ref_seq, patientID)
                        prev_add_chrom = chrom
                        chrom_index += 1
                        chrom = _get_chrom(chrom_index)
                ref_seq = _get_ref_seq_by_chrom(ref_build, record.CHROM)
                if not conversion_region or conversion_region[record.CHROM, record.POS -1: record.POS].empty == False:       
                    _add_record_variants(record, ref_seq, patientID, fhir_helper)

    general_logger.info("Adding all the phased sequence relationship found")
    fhir_helper.add_phased_relationship_obv()

    general_logger.info("Adding all the observation Ids to result block")
    fhir_helper.add_report_result()

    fhir_helper.generate_final_json()
    general_logger.info(f"Export the FHIR object to the file {output_filename}")
    fhir_helper.export_fhir_json(output_filename)
    general_logger.info("Completed conversion")





import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="vcf2fhir",
    version="0.0.12",
    author="",
    test_suite='vcf2fhir.test.test_vcf2fhir.suite',
    author_email="info@elimu.io",
    description="Convert .vcf files to HL7 FHIR standard",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/openelimu/VCF-2-FHIR",
    packages=['vcf2fhir', 'vcf2fhir.test'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: Apache Software License"
    ],
    tests_require = [
    'unittest',
    ],
    install_requires=[
        'Cython >=0.29.21',
        'fhirclient==3.2.0',
        'pysam',
        'pandas',
        'pytz >= 2019.3',
        'pyVCF >=0.6.8',
        'pyranges >= 0.0.79'
    ],
    python_requires='>=3.6',
)


import os
import datetime
import pandas as pd
import pytz
import logging


general_logger = logging.getLogger("vcf2fhir.general")

class _Utilities(object):

    """
    /**
     * generates current date into fhir format 
     * @return date
     */
    """
    def getFhirDate(): 
        z = datetime.datetime.now(pytz.timezone('UTC')).strftime("%Y-%m-%dT%H:%M:%S%z")
        return z[:-2]+':'+z[-2:]

    def getSequenceRelation(phasedRecMap):
        RelationTable= pd.DataFrame(columns=['POS1','POS2','Relation'])
        for key in phasedRecMap:
            prev_record = None
            for record in phasedRecMap[key]:
                if prev_record is None:
                    prev_record = record
                    continue
                prev_data = prev_record.samples[0].data
                record_data = record.samples[0].data
                if(prev_data.PS == record_data.PS):
                    if prev_data.GT == record_data.GT:
                        RelationTable = RelationTable.append({'POS1': prev_record.POS, 'POS2': record.POS, 'Relation': 'Cis'}, ignore_index=True)
                    else:
                        RelationTable = RelationTable.append({'POS1': prev_record.POS, 'POS2': record.POS, 'Relation': 'Trans'}, ignore_index=True)
                prev_record = record
        return RelationTable

    def getAllelicState(record):
        allelicState = ''
        allelicCode = ''
        # Using  the first sample
        sample = record.samples[0]
        alleles = sample.gt_alleles
        if record.CHROM != 'M':
            if len(alleles) >= 2 and sample.gt_type == 1:
                allelicState = 'heterozygous'
                allelicCode = 'LA6706-1'
            elif len(alleles) >= 2:
                allelicState = 'homozygous'
                allelicCode = 'LA6705-3'
            elif sample.gt_type != None and len(alleles) == 1:
                allelicState = 'hemizygous'
                allelicCode = 'LA6707-9'
            else:
                _Utilities._error_log_allelicstate(record)
        elif sample.gt_type != None and len(alleles) == 1 and alleles[0] == '1':
                if hasattr(sample.data, 'AD') and hasattr(sample.data, 'DP'):
                    try:
                        if(isinstance(sample.data.AD, list) and len(sample.data.AD) > 0):
                            ratio = float(sample.data.AD[0])/float(sample.data.DP)
                        else:
                            ratio = float(sample.data.AD)/float(sample.data.DP)
                        if ratio > 0.99:
                            allelicState = "homoplasmic"
                            allelicCode = "LA6704-6"
                        else: 
                            allelicState = "heteroplasmic"
                            allelicCode = "LA6703-8"
                    except Exception as e:
                        general_logger.debug(e)
                        _Utilities._error_log_allelicstate(record)
                        pass
                else:
                    _Utilities._error_log_allelicstate(record)
        else:            
            _Utilities._error_log_allelicstate(record)
        return {'ALLELE': allelicState, 'CODE' : allelicCode}

    def extract_chrom_identifier(chrom):
        chrom = chrom.upper().replace("CHR", "")
        if chrom == "MT":
            chrom = "M"
        return chrom

    def _error_log_allelicstate(record):
            general_logger.error("Cannot Determine AllelicState for: %s , considered sample: %s", record, record.samples[0].data)

        